"""Utility to help read lines from log files"""
import re
from datetime import datetime
from typing import IO
from parse import search
import pandas as pd

CLEAN_MARKUP_REGEX = re.compile('<.*?>')
CLEAN_TIME_AND_TYPE_REGEX = re.compile(r'^(\[\s.*\s\])[ ](\(combat\)[ ])')
CLEAN_CORP_AND_SHIP_REGEX = re.compile(r'\[.*')

class CharacterLogFile:
    """Class representing a character name and log file handle"""
    def __init__(self, filename: str, character: str, file: IO[any]):
        self.filename = filename
        self.character = character
        self.file = file

    def is_open(self) -> bool:
        """Return boolean indicating if file is open"""
        return not self.file.closed

    def close(self):
        """Close the file handle"""
        self.file.close()

def _parse_combat_log_line(line, character):
    parsed_line = search('[ {} ] ({})', line)
    if parsed_line is None:
        return None

    timestamp = datetime.strptime(parsed_line[0], '%Y.%m.%d %H:%M:%S')
    message_type = parsed_line[1]

    if message_type != 'combat':
        return None

    clean_line = re.sub(CLEAN_MARKUP_REGEX, '', line)
    clean_line = re.sub(CLEAN_TIME_AND_TYPE_REGEX, '', clean_line)

    parsed_damge_direction = search('{:d} {:l}', clean_line)

    if parsed_damge_direction is None:
        return None

    if parsed_damge_direction[1] not in ('to', 'from'):
        return None

    damage = parsed_damge_direction[0]
    direction = parsed_damge_direction[1]

    if direction == 'to':
        clip_index = clean_line.find('to') + 3
    elif direction == 'from':
        clip_index = clean_line.find('from') + 5

    subject_what_quality = clean_line[clip_index:].split(' - ')
    if len(subject_what_quality) < 2:
        return None

    subject = subject_what_quality[0]
    subject = re.sub(CLEAN_CORP_AND_SHIP_REGEX, '', subject)
    what = subject_what_quality[1]

    return [
        character,
        timestamp,
        message_type,
        damage,
        direction,
        subject,
        what ]

def _read_and_parse_combat_log_lines(log: CharacterLogFile):
    file = log.file
    character = log.character

    parsed_lines = []

    for line in file.readlines():
        parsed = _parse_combat_log_line(line, character)

        if parsed is not None:
            parsed_lines.append(parsed)

    return parsed_lines

def _flatten(in_list):
    return [item for ll in in_list for item in ll]

def read_all_combat_log_entries(character_files):
    """Read all combat log entries from character log files"""
    if len(character_files) == 0:
        return []

    character_log_entries = map(_read_and_parse_combat_log_lines, character_files)

    return _flatten(character_log_entries)

def open_log_files(
        character_logs,
        filter_characters,
        seek_to_end: bool = False
    ):
    """Open character log files"""
    character_files = []
    for character_log in character_logs:
        name = character_log['character_name']
        path = character_log['path']

        if name not in filter_characters:
            continue

        print(f'open {name} log at {path}')
        file = open(path, 'r', encoding='UTF8')
        if seek_to_end:
            file.seek(0, 2)

        character_files.append(CharacterLogFile(path, name, file))
    return character_files

def log_entries_to_dataframe(log_entries) -> pd.DataFrame:
    """Create DataFrame from log entries"""
    columns = ['character','timestamp','type','damage','direction','subject','what']
    result = pd.DataFrame(log_entries, columns=columns)
    result.sort_values('timestamp', ascending=True)

    return result



"""Functions for finding and filtering log files"""
from datetime import datetime
from os import listdir
from os.path import isfile, join
from pathlib import Path
import pandas as pd
from parse import parse

def _map_log_file(filename: str):
    """Retrieve structured data about log file based on the file name"""
    [name, _] = filename.split('.')
    parts =  name.split('_', 3)

    if len(parts) == 3:
        (date, time, account) = parts
    elif len(parts) == 2:
        (date, time, account) = parts + [None]

    date_time = f'{date} {time}'
    created_at = datetime.strptime(date_time, '%Y%m%d %H%M%S')

    return {
        'filename': filename,
        'account': account,
        'created_at': created_at
    }

def _find_log_files(eve_logs_directory):
    """Find all log files in provided directory"""
    files = [f for f in listdir(eve_logs_directory) if isfile(join(eve_logs_directory, f))]
    files = list(map(_map_log_file, files))

    return pd.DataFrame(files)

LOG_HEADER_SIG = '------------------------------------------------------------'
LOG_HEADER_TITLE = 'Gamelog'

def _read_log_character_name(log_file_path):
    """Read the character name from log header"""
    with open(log_file_path, 'r', encoding='UTF8') as log_file:
        sig = log_file.readline().strip()
        if LOG_HEADER_SIG not in sig:
            return None

        title = log_file.readline().strip()
        if LOG_HEADER_TITLE not in title:
            return None

        character_line = log_file.readline().strip()
        character_name = parse('Listener: {}', character_line)
        if character_name is None:
            return None

    return character_name[0]

def _create_log_result(directory, log):
    log_path = join(directory, log['filename'])
    character_name = _read_log_character_name(log_path)

    return {
        'path': log_path,
        'character_id': log['account'],
        'character_name': character_name,
        'created_at': log['created_at'],
    }

def find_most_recent_character_logs(directory):
    """Find the most recent log file for each account"""
    log_files = _find_log_files(directory)

    log_files.dropna(inplace=True)
    log_files.sort_values(by='created_at', inplace=True, ascending=False)

    unique_accounts = log_files['account'].unique()

    result = []

    for account in unique_accounts:
        account_logs = log_files.loc[log_files['account'] == account]
        if account_logs.count == 0:
            continue

        first_log = account_logs.iloc[0]

        result.append(_create_log_result(directory, first_log))

    return result

def find_character_logs_at_date(directory, date):
    """Find all log files at date"""
    log_files = _find_log_files(directory)
    log_files.dropna(inplace=True)

    log_files['date'] = log_files['created_at'].apply(_date_string)

    logs_at_date = log_files['date'].isin([date])
    filtered_logs = log_files[logs_at_date]

    return list(
        filtered_logs.apply(lambda log: _create_log_result(directory, log), axis=1)
    )

def _date_string(date: datetime):
    return date.date().strftime('%Y-%m-%d')

def find_all_dates(directory):
    """Find all available log file dates"""
    log_files = _find_log_files(directory)
    log_files.dropna(inplace=True)

    return list(log_files['created_at'].apply(_date_string).unique())

def default_log_directory():
    """Retrieve the default Eve log directory"""
    user_directory = join(Path.home(), 'Documents\\EVE\\logs\\Gamelogs')

    return user_directory



"""Configuration UI"""
import gradio as gr

from smarterbombing.configuration import save_configuration

def render_configuration(configuration, configuration_path):
    """Render the configuration UI"""
    def _character_list():
        return ', '.join(configuration['characters'])

    def _add_character(character_name):
        characters = configuration['characters']
        characters.append(character_name)
        save_configuration(configuration, configuration_path)
        print(f'added character -> {character_name}')

        return _character_list()

    def _change_log_directory(new_log_directory):
        configuration['log_directory'] = new_log_directory
        save_configuration(configuration, configuration_path)
        print(f'changed log_directory -> {new_log_directory}')

    def _change_dps_rolling_window(seconds):
        configuration['dps_rolling_window_seconds'] = seconds
        save_configuration(configuration, configuration_path)

    dps_rolling_window_seconds = gr.Slider(label='Average Window (Seconds)',
                                           minimum=0.0, maximum=300,
                                           value=configuration['dps_rolling_window_seconds'])
    dps_rolling_window_seconds.change(fn=_change_dps_rolling_window,
                                      inputs=[dps_rolling_window_seconds])

    log_directory = gr.Textbox(
        value=configuration['log_directory'],
        label="EVE Online Log Directory")
    log_directory.submit(fn=_change_log_directory, inputs=[log_directory])

    add_character = gr.Textbox(
        placeholder='Character name',
        label='Add Character')

    gr.Markdown('### Characters')
    character_list = gr.Markdown(_character_list())

    add_character.submit(fn=_add_character, inputs=[add_character], outputs=[character_list])



"""Configuration helpers"""
import json
from os.path import isfile
from smarterbombing.logs import default_log_directory

def load_configuration(configuration_path):
    """Load configuration"""

    if not isfile(configuration_path):
        configuration = {
            'log_directory': default_log_directory(),
            'characters': [],
            'dps_rolling_window_seconds': 10,
        }

        save_configuration(configuration, configuration_path)

        return configuration

    with open(configuration_path, 'rt', encoding='UTF8') as config_file:
        return json.load(config_file)
    
def save_configuration(configuration, configuration_path):
    """Save configuration"""
    with open(configuration_path, 'wt', encoding='UTF8') as config_file:
        json.dump(configuration, config_file)


"""Offline analysis UI"""
import gradio as gr
from smarterbombing.analysis import average_dps_per_character_melt
from smarterbombing.app_offline import AppOffline

from smarterbombing.logs import find_all_dates

def render_offline_analysis(app: AppOffline, configuration):
    """Render the offline analysis UI"""

    def _damage_over_time_hostile():
        try:
            data = app.damage_over_time_hostile()
            if data.empty:
                return None
            
            data = average_dps_per_character_melt(data)
        except Exception as exc:
            raise gr.Error('Couldn\'t process data. Perhaps there\'s no session on selected date.') from exc
        
        return data

    def _damage_over_time_friendly():
        try:
            data = app.damage_over_time_friendly()
            if data.empty:
                return None
            
            data = average_dps_per_character_melt(data)
        except Exception as exc:
            raise gr.Error('Couldn\'t process data. Perhaps there\'s no session on selected date.') from exc

        return data

    def _damage_over_time_incoming_hostile():
        try:
            data = app.damage_over_time_incoming_hostile()
            if data.empty:
                return None
            
            data = average_dps_per_character_melt(data)
        except Exception as exc:
            raise gr.Error('Couldn\'t process data. Perhaps there\'s no session on selected date.') from exc

        return data

    def _damage_over_time_incoming_friendly():
        try:
            data = app.damage_over_time_incoming_friendly()
            if data.empty:
                return None
            
            data = average_dps_per_character_melt(data)
        except Exception as exc:
            raise gr.Error('Couldn\'t process data. Perhaps there\'s no session on selected date.') from exc

        return data

    with gr.Column():
        with gr.Row():
            gr.Markdown("""### Offline Analysis
            Look back at and analyse performance from old log files. This can be used to 
            compare statistics from various ratting sessions. Select a date from the drop
            down menu on the right and wait a moment for data to be crunched by the system.""")

            log_dates = find_all_dates(configuration['log_directory'])
            selected_date = gr.Dropdown(choices=log_dates, label='Date')

        with gr.Accordion(label='Sessions'):
            sessions = gr.DataFrame(
                label='Sessions',
                headers=['Date', 'Start', 'End'],
                interactive=False)

    with gr.Column():
        with gr.Row():
            dps_out_h = gr.LinePlot(
                x_title='Time',
                x='timestamp',
                y_title='DPS',
                y='damage',
                color='character',
                title='Outgoing Damage',
                width=530)

            dps_out_f = gr.LinePlot(
                x_title='Time',
                x='timestamp',
                y_title='DPS',
                y='damage',
                color='character',
                title='Outgoing Damage (Friendly)',
                width=530)
        with gr.Row():
            dps_in_h = gr.LinePlot(
                x_title='Time',
                x='timestamp',
                y_title='DPS',
                y='damage',
                color='character',
                title='Incoming Damage',
                width=530)

            dps_in_f = gr.LinePlot(
                x_title='Time',
                x='timestamp',
                y_title='DPS',
                y='damage',
                color='character',
                title='Incoming Damage (Friendly)',
                width=530)

    selected_date.change(
        app.load_at_date,
        inputs=[selected_date],
        outputs=[sessions],
    ).then(
        _damage_over_time_hostile,
        outputs=dps_out_h
    ).then(
        _damage_over_time_friendly,
        outputs=dps_out_f
    ).then(
        _damage_over_time_incoming_hostile,
        outputs=dps_in_h
    ).then(
        _damage_over_time_incoming_friendly,
        outputs=dps_in_f
    )



"""Live UI"""
import gradio as gr
import pandas as pd
from tabulate import tabulate
from smarterbombing.analysis import average_dps_per_character_melt

from smarterbombing.app_live import AppLive

def render_live(sb_ui: gr.Blocks, app: AppLive):
    """Render live UI"""

    def _update_app():
        app.update()

    def _update_log_timestamp_text():
        return gr.update(value=f'Log Timestamp: {app.most_recent_timestamp}')

    def _update_open_log_files():
        if not app.is_logs_open():
            return gr.update(visible=False)

        log_table = list(map(
            lambda log: { 'File': log.filename, 'Character': log.character },
            app.logs
        ))

        return gr.update(value=pd.DataFrame(log_table), visible=True)

    def _update_outgoing_hostile_dps():
        if app.outgoing_hostile_damage.empty:
            return gr.update(visible=False)

        value = average_dps_per_character_melt(app.outgoing_hostile_damage)

        return gr.update(value=value, visible=True)

    def _update_outgoing_friendly_dps():
        if app.outgoing_friendly_damage.empty:
            return gr.update(visible=False)

        value = average_dps_per_character_melt(app.outgoing_friendly_damage)

        return gr.update(value=value, visible=True)

    def _update_incoming_hostile_dps():
        if app.incoming_hostile_damage.empty:
            return gr.update(visible=False)

        value = average_dps_per_character_melt(app.incoming_hostile_damage)

        return gr.update(value=value, visible=True)

    def _update_incoming_friendly_dps():
        if app.incoming_friendly_damage.empty:
            return gr.update(visible=False)

        value = average_dps_per_character_melt(app.incoming_friendly_damage)

        return gr.update(value=value, visible=True)

    def _update_site_statistics():
        app.update_site_statistics()
        value = app.site_statistics

        if value.empty:
            return gr.update(visible=False)

        value = tabulate(value, tablefmt='html', headers='keys')

        return gr.update(value=value, visible=True)

    def _update_compound_site_statistics():
        app.update_compound_site_statistics()
        value = app.site_compound_statistics

        if value.empty:
            return gr.update(visible=False)

        value = tabulate(value, tablefmt='html', headers='keys')

        return gr.update(value=value, visible=True)

    def _reload_files():
        app.open_logs()

        return _update_open_log_files()

    def _close_files():
        app.close_logs()

        return _update_open_log_files()

    def _clear_data():
        app.clear_data()

    with gr.Column():
        with gr.Row():
            dps_out_h = gr.LinePlot(
                x_title='Time',
                x='timestamp',
                y_title='DPS',
                y='damage',
                color='character',
                title='Outgoing Damage',
                interactive=False,
                visible=False,
                width=530)

            dps_in_h = gr.LinePlot(
                x_title='Time',
                x='timestamp',
                y_title='DPS',
                y='damage',
                color='character',
                title='Incoming Damage',
                interactive=False,
                visible=False,
                width=530)

        with gr.Row():
            dps_out_f = gr.LinePlot(
                x_title='Time',
                x='timestamp',
                y_title='DPS',
                y='damage',
                color='character',
                title='Outgoing Damage (Friendly)',
                interactive=False,
                visible=False,
                width=530)

            dps_in_f = gr.LinePlot(
                x_title='Time',
                x='timestamp',
                y_title='DPS',
                y='damage',
                color='character',
                title='Incoming Damage (Friendly)',
                interactive=False,
                visible=False,
                width=530)

        with gr.Column():
            compound_site_statistics = gr.HTML()
            site_statistics = gr.HTML()

    gr.Markdown(value='_Graphs and statistics will appear once started ratting._')

    with gr.Column():
        with gr.Accordion('Diagnostics', open=False):
            with gr.Row():
                reload_files = gr.Button(value='Reload Files')
                close_files = gr.Button(value='Close Files')
                clear_data = gr.Button(value='Clear Data')

            log_timestamp = gr.Markdown(value='Most recent log timestamp: None')
            open_logfiles = gr.Dataframe(value=[], label='Open Logs', visible=False)

            reload_files.click(fn=_reload_files, outputs=open_logfiles)
            close_files.click(fn=_close_files, outputs=open_logfiles)
            clear_data.click(fn=_clear_data)

    sb_ui.load(_update_app, None, None, every=1)
    sb_ui.load(_update_log_timestamp_text, None, log_timestamp, every=5)
    sb_ui.load(_update_open_log_files, None, open_logfiles, every=5)

    sb_ui.load(_update_outgoing_hostile_dps, None, dps_out_h, every=1)
    sb_ui.load(_update_outgoing_friendly_dps, None, dps_out_f, every=1)
    sb_ui.load(_update_incoming_hostile_dps, None, dps_in_h, every=1)
    sb_ui.load(_update_incoming_friendly_dps, None, dps_in_f, every=1)
    sb_ui.load(_update_site_statistics, None, site_statistics, every=5)
    sb_ui.load(_update_compound_site_statistics, None, compound_site_statistics, every=5)



"""Analyze logs offlinee"""
from datetime import datetime, timedelta
import pandas as pd
from gradio import Progress

from smarterbombing.logs import find_character_logs_at_date
from smarterbombing.log_reader import\
    read_all_combat_log_entries,\
    open_log_files,\
    log_entries_to_dataframe

EVENT_GROUP_ALL_DAMAGE='all_combat_events'
EVENT_GROUP_OUTGOING_DAMAGE='outgoing_damage'
EVENT_GROUP_OUTGOING_HOSTILE_DAMAGE='outgoing_hostile_damage'
EVENT_GROUP_OUTGOING_FRIENDLY_DAMAGE='outgoing_friendly_damage'
EVENT_GROUP_INCOMING_DAMAGE='incoming_damage'
EVENT_GROUP_INCOMING_HOSTILE_DAMAGE='incoming_hostile_damage'
EVENT_GROUP_INCOMING_FRIENDLY_DAMAGE='incoming_friendly_damage'

def _filter_by_direction(events: pd.DataFrame, direction: str):
    filter_outgoing = events['direction'].isin([direction])
    return events[filter_outgoing]

def _filter_unfriendly_damage(events: pd.DataFrame, characters):
    filter_friendly = events['subject'].isin(characters)
    return events[~filter_friendly]

def _filter_friendly_damage(events: pd.DataFrame, characters):
    filter_friendly = events['subject'].isin(characters)
    return events[filter_friendly]

def _group_by_delta_time(events: pd.DataFrame, max_gap_seconds: float):
    events['delta'] = events['timestamp'].diff() > timedelta(seconds=max_gap_seconds)

    groups = []
    for _, group in events.groupby([events['delta'].cumsum()]):
        groups.append(group)

    return groups

def group_damage_events(events: pd.DataFrame, characters):
    """Sort combat events by type and direction"""
    outgoing_damage = _filter_by_direction(events, 'to')
    outgoing_damage_to_friendly = _filter_friendly_damage(outgoing_damage, characters)
    outgoing_damage_to_hostile = _filter_unfriendly_damage(outgoing_damage, characters)

    incoming_damage = _filter_by_direction(events, 'from')
    incoming_damage_from_friendly = _filter_friendly_damage(incoming_damage, characters)
    incoming_damage_from_hostile = _filter_unfriendly_damage(incoming_damage, characters)

    return {
        EVENT_GROUP_ALL_DAMAGE: events,
        EVENT_GROUP_OUTGOING_DAMAGE: outgoing_damage,
        EVENT_GROUP_OUTGOING_HOSTILE_DAMAGE: outgoing_damage_to_hostile,
        EVENT_GROUP_OUTGOING_FRIENDLY_DAMAGE: outgoing_damage_to_friendly,
        EVENT_GROUP_INCOMING_DAMAGE: incoming_damage,
        EVENT_GROUP_INCOMING_HOSTILE_DAMAGE: incoming_damage_from_hostile,
        EVENT_GROUP_INCOMING_FRIENDLY_DAMAGE: incoming_damage_from_friendly,
    }

def parse_logs(configuration, date, progress=Progress()):
    """Parse logs collecting events and best effort split into sessions"""

    log_directory = configuration['log_directory']
    characters = configuration['characters']

    progress(0, desc='Locating log files')
    character_logs = find_character_logs_at_date(log_directory, date)

    progress(0.1, desc='Opening log files')
    character_files = open_log_files(character_logs, characters)

    progress(0.2, desc='Parsing log messages')
    combat_log_entries = list(read_all_combat_log_entries(character_files))
    combat_log_entries = log_entries_to_dataframe(combat_log_entries)

    sessions = _group_by_delta_time(combat_log_entries, 350)

    info = []
    data = []

    for session in sessions:
        info.append({
            'Date': date,
            "Start": session.iloc[0]['timestamp'].strftime('%H:%M:%S'),
            "End": session.iloc[-1]['timestamp'].strftime('%H:%M:%S'),
            "Events": len(session.index),
        })

        grouped_events = group_damage_events(session, characters)

        data.append(grouped_events)

    progress(1.0, desc='Done')

    return (data, pd.DataFrame(info))

def reshape_for_per_character_dps(data: pd.DataFrame) -> pd.DataFrame:
    """Reshape dataframe to represent damage per character"""
    data = data[['timestamp', 'character', 'damage']]
    data = data.groupby(['timestamp', 'character']).sum().reset_index()
    data = data.pivot(
        index='timestamp',
        columns='character',
        values='damage'
    ).fillna(0.0)

    return data

def fixed_window_average_dps_per_character(
        data: pd.DataFrame,
        template: pd.DataFrame,
        start_at: datetime,
        end_at: datetime,
        average_seconds: int = 10) -> pd.DataFrame:
    """Create a fixed time window and populate with average DPS"""

    fixed_window = pd.concat([
        generate_1s_dataframe(start_at, end_at),
        template,
    ]).fillna(0.0)\
        .set_index('timestamp')\
        .rename_axis('character', axis='columns')

    data = reshape_for_per_character_dps(data)
    data = data.combine_first(fixed_window)

    data = data.assign(Total=data.sum(1))
    data = data.resample('1S').asfreq(fill_value=0.0)

    if average_seconds > 0:
        data = data.rolling(timedelta(seconds=average_seconds)).mean()

    return data

def site_statistics(data: pd.DataFrame, characters, minimum_gap_seconds: int = 30) -> pd.DataFrame:
    """Calculate site based time statistics"""
    data = _filter_by_direction(data, 'to')
    data = data.sort_values(by='timestamp')
    data = data.reset_index()

    # Calculate event deltas
    data['delta_time'] = data['timestamp'].diff()

    # Assign site index based on time
    data['new_site'] = data['delta_time'] > timedelta(seconds=minimum_gap_seconds)
    data['site_index'] = data['new_site'].cumsum()

    if not data['new_site'].any():
        return create_empty_site_statistics_dataframe()

    mask_friendly = data['subject'].apply(lambda c: c in characters)
    data['hit_friendly'] = mask_friendly

    data_no_friendly = data[~data['hit_friendly']]

    filter_new_sites = data['new_site']
    site_downtime = data[filter_new_sites][['site_index', 'delta_time']]\
        .set_index('site_index')\
        .rename(columns={'delta_time': 'downtime'})

    site_groups = data.groupby('site_index')
    start_times = site_groups['timestamp'].first()
    end_times = site_groups['timestamp'].last()

    site_groups = data_no_friendly.groupby('site_index')
    effective_start_times = site_groups['timestamp'].first()
    effective_end_times = site_groups['timestamp'].last()

    site_damage = data_no_friendly.pivot(columns='site_index', values='damage').sum()
    site_hits = data_no_friendly.pivot(columns='site_index', values='subject').count()

    durations = end_times - start_times
    effective_durations = effective_end_times - effective_start_times

    result = pd.DataFrame({
        'effective_start_time': effective_start_times,
        'effective_end_time': effective_end_times,
        'start_time': start_times,
        'end_time': end_times,
        'duration': durations,
        'effective_duration': effective_durations,
        'damage': site_damage,
        'hits': site_hits,
    })

    result = result.combine_first(site_downtime).fillna(timedelta(0))

    return result

def compound_site_statistics(data: pd.DataFrame):
    """Calculate compound site statistics"""
    total_time = data.iloc[-1]['end_time'] - data.iloc[0]['start_time']

    total_seconds = total_time.total_seconds()
    total_hours = total_seconds / 3600

    total_downtime = data['downtime'].sum()

    average_downtime = data['downtime'].mean()
    average_site_time = data['duration'].mean()

    average_effective_time = data['effective_duration'].mean()
    effective_time = data['effective_duration'].sum()

    effective_seconds = effective_time.total_seconds()
    time_efficiency = effective_seconds / total_seconds

    sites_per_hour = len(data.index) / total_hours

    total_site_damage = data['damage'].sum()
    average_site_damage = data['damage'].mean()

    average_damage_per_second = total_site_damage / effective_seconds

    return pd.DataFrame([{
        'total_time': total_time,
        'total_downtime': total_downtime,
        'total_effective_time': effective_time,
        'sites_per_hour': sites_per_hour,
        'average_downtime': average_downtime,
        'average_time': average_site_time,
        'average_effective_time': average_effective_time,
        'average_damage': average_site_damage,
        'total_damage': total_site_damage,
        'damage_per_second': average_damage_per_second,
        'time_efficiency': time_efficiency
    }])

def average_dps_per_character(data: pd.DataFrame, average_seconds: int = 10) -> pd.DataFrame:
    """Calculate average DPS per character"""

    if data.empty:
        return data

    data = reshape_for_per_character_dps(data)

    data = data.assign(Total=data.sum(1))
    data = data.resample('1S').asfreq(fill_value=0.0)

    if average_seconds > 0:
        data = data.rolling(timedelta(seconds=average_seconds)).mean()

    return data

def average_dps_per_character_melt(data: pd.DataFrame) -> pd.DataFrame:
    """Reshape DataFrame to long format"""
    data = data.reset_index().melt(id_vars='timestamp', value_name='damage')

    return data

def resample_30s_mean(data: pd.DataFrame) -> pd.DataFrame:
    """Resample data to 30s average rows"""

    return data.resample('30S').mean()

def generate_1s_dataframe(start_date: datetime, end_date: datetime) -> pd.DataFrame:
    """Generate index of datetime range with 1 second frequency"""
    start_date = start_date.replace(microsecond=0)
    end_date = end_date.replace(microsecond=0)

    timerange = pd.date_range(start_date, end_date, freq='S', unit='s')

    return timerange.to_frame(index=False, name='timestamp')

def filter_by_datetime(data: pd.DataFrame, start_date: datetime, end_date: datetime):
    """Return rows which timestamp is within the provided datetimes"""
    return data[(data['timestamp'] >= start_date) & (data['timestamp'] <= end_date)]

def create_empty_damage_dataframe():
    """Create an empty damage DataFrame"""
    return pd.DataFrame(columns=['timestamp', 'character', 'damage'])

def create_empty_site_statistics_dataframe():
    """Create an empty site statistics DataFrame"""
    return pd.DataFrame(columns=[
        'effective_start_time',
        'effective_end_time',
        'start_time',
        'end_time',
        'duration',
        'effective_duration',
        'damage',
        'hits',
    ])

def create_empty_compound_site_statistics_dataframe():
    """Create an empty compound site statistics DataFrame"""
    return pd.DataFrame(columns=[
        'total_time',
        'total_downtime',
        'total_effective_time',
        'sites_per_hour',
        'average_downtime',
        'average_time',
        'average_effective_time',
        'average_damage',
        'total_damage',
        'damage_per_second',
        'time_efficiency',
    ])



"""Offline application"""
import pandas as pd
from gradio import Progress
from smarterbombing.analysis import parse_logs, average_dps_per_character, resample_30s_mean

class AppOffline:
    """Class holding state for offline processing"""
    def __init__(self, configuration) -> None:
        self.configuration = configuration
        self.session_index = 0
        self.data = None
        self.info = None

    def load_at_date(self, date, progress=Progress()):
        """Load logs at date"""
        (data, info) = parse_logs(self.configuration, date, progress)

        self.info = info
        self.data = data

        return info

    def select_session(self, session_index):
        """Select session with index"""
        self.session_index = session_index

    def selected_session(self) -> str:
        """Returns a list of session choices"""
        if not self.loaded():
            return None
        return self.info['Start'].iloc[0]

    def damage_over_time_hostile(self) -> pd.DataFrame:
        """Returns a dataframe with DPS"""
        if not self.loaded():
            return pd.DataFrame([])

        session = self.data[self.session_index]
        data = session['outgoing_hostile_damage']

        if (data.empty):
            return pd.DataFrame([])

        average_dps = average_dps_per_character(data)
        average_dps = resample_30s_mean(average_dps)

        return average_dps

    def damage_over_time_friendly(self) -> pd.DataFrame:
        """Returns a dataframe with DPS"""
        if not self.loaded():
            return pd.DataFrame([])

        session = self.data[self.session_index]
        data = session['outgoing_friendly_damage']

        if (data.empty):
            return pd.DataFrame([])

        average_dps = average_dps_per_character(data)
        average_dps = resample_30s_mean(average_dps)

        return average_dps

    def damage_over_time_incoming_hostile(self) -> pd.DataFrame:
        """Returns a dataframe with DPS"""
        if not self.loaded():
            return pd.DataFrame([])

        session = self.data[self.session_index]
        data = session['incoming_hostile_damage']

        if (data.empty):
            return pd.DataFrame([])

        average_dps = average_dps_per_character(data)
        average_dps = resample_30s_mean(average_dps)

        return average_dps

    def damage_over_time_incoming_friendly(self) -> pd.DataFrame:
        """Returns a dataframe with DPS"""
        if not self.loaded():
            return pd.DataFrame([])

        session = self.data[self.session_index]
        data = session['incoming_friendly_damage']

        if (data.empty):
            return pd.DataFrame([])

        average_dps = average_dps_per_character(data)
        average_dps = resample_30s_mean(average_dps)

        return average_dps

    def loaded(self) -> bool:
        """Return boolean indicating if data is loaded"""
        return self.info is not None and self.data is not None



"""User interface"""
import gradio as gr

from smarterbombing.app_live import AppLive
from smarterbombing.app_offline import AppOffline
from smarterbombing.configuration import load_configuration
from smarterbombing.webui.ui_configuration import render_configuration
from smarterbombing.webui.ui_live import render_live
from smarterbombing.webui.ui_offline_analysis import render_offline_analysis

CONFIGURATION_PATH = 'configuration.json'

def run_webui(port: int = 42069):
    """Run web interface"""
    configuration = load_configuration(CONFIGURATION_PATH)

    app_offline = AppOffline(configuration)
    app_live = AppLive(configuration)
    app_live.open_logs()

    with gr.Blocks(title="Smarterbombing") as sb_ui:
        with gr.Tab('Live'):
            render_live(sb_ui, app_live)

        with gr.Tab('Offline Analysis'):
            render_offline_analysis(app_offline, configuration)

        with gr.Tab('Configuration'):
            render_configuration(configuration, CONFIGURATION_PATH)

    sb_ui.queue()
    sb_ui.launch(show_api=False, server_port=port)

    app_live.close_logs()



"""Live application"""
from datetime import timedelta, datetime, timezone
import pandas as pd
from smarterbombing.analysis import (
    EVENT_GROUP_INCOMING_FRIENDLY_DAMAGE,
    EVENT_GROUP_INCOMING_HOSTILE_DAMAGE,
    EVENT_GROUP_OUTGOING_FRIENDLY_DAMAGE,
    EVENT_GROUP_OUTGOING_HOSTILE_DAMAGE,
    compound_site_statistics,
    create_empty_compound_site_statistics_dataframe,
    create_empty_damage_dataframe,
    create_empty_site_statistics_dataframe,
    filter_by_datetime,
    fixed_window_average_dps_per_character,
    group_damage_events,
    site_statistics)

from smarterbombing.logs import find_most_recent_character_logs
from smarterbombing.log_reader import\
    CharacterLogFile,\
    log_entries_to_dataframe,\
    open_log_files,\
    read_all_combat_log_entries

class AppLive:
    """Live application"""
    def __init__(self, configuration):
        self.configuration = configuration
        self.current_time = datetime.now(timezone.utc)

        self.logs = []

        self.data = pd.DataFrame([])
        self.outgoing_hostile_damage = create_empty_damage_dataframe()
        self.outgoing_friendly_damage = create_empty_damage_dataframe()
        self.incoming_hostile_damage = create_empty_damage_dataframe()
        self.incoming_friendly_damage = create_empty_damage_dataframe()

        self.site_statistics = create_empty_site_statistics_dataframe()
        self.site_compound_statistics = create_empty_compound_site_statistics_dataframe()

        self.damage_data_template = pd.DataFrame(columns=configuration['characters'])\
            .rename_axis('character', axis='columns')

        self.data_rows = 0
        self.most_recent_timestamp = None
        self.damage_graph_time_window = timedelta(minutes=5)

    def open_logs(self):
        """Open log files"""
        most_recent_logs = find_most_recent_character_logs(self.configuration['log_directory'])

        for log in self.logs:
            log.close()

        self.logs = open_log_files(most_recent_logs, self.configuration['characters'], True)

    def close_logs(self):
        """Close all log files"""
        for log in self.logs:
            log.close()

        self.logs = []

    def clear_data(self):
        """Clear all data"""
        self.data = pd.DataFrame([])

        self.outgoing_hostile_damage = create_empty_damage_dataframe()
        self.outgoing_friendly_damage = create_empty_damage_dataframe()
        self.incoming_hostile_damage = create_empty_damage_dataframe()
        self.incoming_friendly_damage = create_empty_damage_dataframe()

        self.site_statistics = create_empty_site_statistics_dataframe()
        self.site_compound_statistics = create_empty_compound_site_statistics_dataframe()

    def is_logs_open(self) -> bool:
        """Return boolean indicating if any log files is open"""
        return any(map(lambda log: log.is_open(), self.logs))

    def parse_new_log_messages(self):
        """Parse all new log messages"""
        parsed_logs = read_all_combat_log_entries(self.logs)

        if len(parsed_logs) > 0:
            parsed_logs = log_entries_to_dataframe(parsed_logs)
            self.data = pd.concat([self.data, parsed_logs])

    def recalculate_damage_graphs(self, graph_from: datetime, graph_until: datetime):
        """Recalculate average dps data"""
        damage_graph_data = filter_by_datetime(self.data, graph_from, graph_until)
        grouped_events = group_damage_events(damage_graph_data, self.configuration['characters'])
        average_seconds = self.configuration['dps_rolling_window_seconds']

        self.outgoing_hostile_damage = fixed_window_average_dps_per_character(
            grouped_events[EVENT_GROUP_OUTGOING_HOSTILE_DAMAGE],
            self.damage_data_template, graph_from, graph_until, average_seconds)
        self.outgoing_friendly_damage = fixed_window_average_dps_per_character(
            grouped_events[EVENT_GROUP_OUTGOING_FRIENDLY_DAMAGE],
            self.damage_data_template, graph_from, graph_until, average_seconds)
        self.incoming_hostile_damage = fixed_window_average_dps_per_character(
            grouped_events[EVENT_GROUP_INCOMING_HOSTILE_DAMAGE],
            self.damage_data_template, graph_from, graph_until, average_seconds)
        self.incoming_friendly_damage = fixed_window_average_dps_per_character(
            grouped_events[EVENT_GROUP_INCOMING_FRIENDLY_DAMAGE],
            self.damage_data_template, graph_from, graph_until, average_seconds)

    def update(self):
        """Read logs and recalculate derived data"""
        self.current_time = datetime.now(timezone.utc).replace(tzinfo=None)

        self.parse_new_log_messages()

        self.data_rows = len(self.data.index)

        if self.data_rows > 0:
            self.most_recent_timestamp = self.data['timestamp'].max()

            damage_graph_from = self.current_time - self.damage_graph_time_window
            damage_graph_until = self.current_time
            self.recalculate_damage_graphs(damage_graph_from, damage_graph_until)

    def update_site_statistics(self):
        """Update site statistics"""
        if not self.data.empty:
            self.site_statistics = site_statistics(self.data, self.configuration['characters'])

    def update_compound_site_statistics(self):
        """Update compound site statistics"""
        if not self.site_statistics.empty:
            self.site_compound_statistics = compound_site_statistics(self.site_statistics)
